import logging
import warnings

import torch

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, TensorLike
from autoemulate.data.utils import set_random_seed
from autoemulate.emulators.base import ProbabilisticEmulator
from autoemulate.simulations.base import Simulator

logger = logging.getLogger("autoemulate")


class HistoryMatching(TorchDeviceMixin):
    r"""
    History Matching class for model calibration.

    History matching is a model calibration method, which uses observed data to
    rule out ``implausible`` parameter values. The implausibility metric is:

    .. math::
        I_i(\bar{x_0}) = \frac{|z_i - \\mathbb{E}(f_i(\bar{x_0}))|}
        {\\sqrt{\text{Var}[z_i - \\mathbb{E}(f_i(\bar{x_0}))]}}

    Queried parameters above a given implausibility threshold are ruled out (RO)
    whereas all other parameters are marked as not ruled out yet (NROY).
    """

    def __init__(
        self,
        observations: dict[str, tuple[float, float]] | dict[str, float],
        threshold: float = 3.0,
        model_discrepancy: float = 0.0,
        rank: int = 1,
        device: DeviceLike | None = None,
    ):
        """
        Initialize the history matching object.

        Parameters
        ----------
        observations: dict[str, tuple[float, float] | dict[str, float]
            For each output variable, specifies observed [value, noise]. In case
            of no uncertainty in observations, provides just the observed value.
        threshold: float
            Implausibility threshold (query points with implausibility scores that
            exceed this value are ruled out). Defaults to 3, which is considered
            a good value for simulations with a single output.
        model_discrepancy: float
            Additional variance to include in the implausibility calculation.
        rank: int
            Scoring method for multi-output problems. Must be 1 <= rank <= n_outputs.
            When the implausibility scores are ordered across outputs, it indicates
            which rank to use when determining whether the query point is NROY. The
            default of ``1`` indicates that the largest implausibility will be used.
        device: DeviceLike | None
            The device to use. If None, the default torch device is returned.
        """
        TorchDeviceMixin.__init__(self, device=device)

        self.threshold = threshold
        self.discrepancy = model_discrepancy
        self.out_dim = len(observations)

        if rank > self.out_dim or rank < 1:
            raise ValueError(
                f"Rank ({rank}) is outside valid range between 1 and output dimension "
                f"of simulator ({self.out_dim})",
            )
        self.rank = rank

        # Save mean and variance of observations, shape: [1, n_outputs]
        self.obs_means, self.obs_vars = self._process_observations(observations)

    def _process_observations(
        self,
        observations: dict[str, tuple[float, float]] | dict[str, float],
    ) -> tuple[TensorLike, TensorLike]:
        """
        Turn observations into tensors of shape [1, n_inputs].

        Parameters
        ----------
        observations: dict[str, tuple[float, float] | dict[str, float]
            For each output variable, specifies observed [value, noise]. In case
            of no uncertainty in observations, provides just the observed value.

        Returns
        -------
        tuple[TensorLike, TensorLike]
            Tensors of observations and the associated noise (which can be 0).
        """
        values = torch.tensor(list(observations.values()), device=self.device)

        # No variance
        if values.ndim == 1:
            means = values
            variances = torch.zeros_like(means, device=self.device)
        # Values are (mean, variance)
        elif values.ndim == 2:
            means = values[:, 0]
            variances = values[:, 1]
        else:
            msg = "Observations must be either float or tuple of two floats."
            raise ValueError(msg)

        # Reshape observation tensors for broadcasting
        return means.view(1, -1), variances.view(1, -1)

    def _create_nroy_mask(self, implausibility: TensorLike) -> TensorLike:
        """
        Create mask for NROY points based on rank.

        Parameters
        ----------
        implausibility: TensorLike
            Tensor of implausibility scores for tested parameters.

        Returns
        -------
        TensorLike
            Tensor indicating whether each implausability score is NROY
            given self.rank and self.threshold values.
        """
        # Sort implausibilities for each sample (descending)
        I_sorted, _ = torch.sort(implausibility, dim=1, descending=True)
        # The rank-th highest output implausibility must be <= threshold
        return I_sorted[:, self.rank - 1] <= self.threshold

    def get_nroy(
        self, implausibility: TensorLike, x: TensorLike | None = None
    ) -> TensorLike:
        """
        Get indices of NROY points from implausibility scores.

        If `x` is provided, returns parameter values at NROY indices.

        Parameters
        ----------
        implausibility: TensorLike
            Tensor of implausibility scores for tested input parameters.
        x: Tensorlike | None
            Optional tensor of scored input parameters.

        Returns
        -------
        TensorLike
            Indices of NROY points or `x` parameters at NROY indices.
        """
        nroy_mask = self._create_nroy_mask(implausibility)
        idx = torch.where(nroy_mask)[0]
        if x is None:
            return idx
        return x[idx]

    def get_ro(
        self, implausibility: TensorLike, x: TensorLike | None = None
    ) -> TensorLike:
        """
        Get indices of RO points from implausibility scores.

        If `x` is provided, returns parameter values at RO indices.

        Parameters
        ----------
        implausibility: TensorLike
            Tensor of implausibility scores for tested input parameters.
        x: Tensorlike | None
            Optional tensor of scored iput parameters.

        Returns
        -------
        TensorLike
            Indices of RO points or `x` parameters at RO indices.
        """
        nroy_mask = self._create_nroy_mask(implausibility)
        idx = torch.where(~nroy_mask)[0]
        if x is None:
            return idx
        return x[idx]

    def calculate_implausibility(
        self,
        pred_means: TensorLike,  # [n_samples, n_outputs]
        pred_vars: TensorLike,  # [n_samples, n_outputs]
    ) -> TensorLike:
        """
        Calculate implausibility scores.

        Parameters
        ----------
        pred_means: TensorLike
            Tensor of prediction means [n_samples, n_outputs]
        pred_vars: TensorLike
            Tensor of prediction variances [n_samples, n_outputs].

        Returns
        -------
        TensorLike
            Tensor of implausibility scores.
        """
        # Additional variance due to model discrepancy (defaults to 0)
        discrepancy = torch.full_like(
            self.obs_vars, self.discrepancy, device=self.device
        )

        # Calculate total variance
        Vs = pred_vars + discrepancy + self.obs_vars

        # Calculate implausibility
        return torch.abs(self.obs_means - pred_means) / torch.sqrt(Vs)

    @staticmethod
    def generate_param_bounds(
        nroy_x: TensorLike,
        buffer_ratio: float = 0.05,
        param_names: list[str] | None = None,
        min_samples: int = 1,
    ) -> dict[str, tuple[float, float]] | None:
        """
        Generate lower/upper parameter bounds as min/max of NROY samples.

        Parameters
        ----------
        nroy_x: TensorLike
            A tensor of NROY parameter samples [n_samples, n_inputs].
        buffer_ratio: float
            A scaling factor used to expand the bounds of the (NROY) parameter space.
            It is applied as a ratio of the range (max_val - min_val) of each input
            parameter to create a buffer around the NROY minimum and maximum values.
        param_names: list[str] | None
            Optional list of parameter names. If None, uses default `["x1", ..., "xn"]`.
        min_samples: int
            Minimum number of samples needed to generate new bounds.

        Returns
        -------
        dict[str, [float, float]] | None
            The generated [lower, upper] parameter bounds. Returns None if there are
            not enough samples to generate bounds from.
        """
        if param_names is None:
            param_names = [f"x{i + 1}" for i in range(nroy_x.shape[1])]

        if nroy_x.shape[0] > min_samples:
            min_val = torch.min(nroy_x, dim=0).values
            max_val = torch.max(nroy_x, dim=0).values
            buffer = (max_val - min_val) * buffer_ratio
            lower_bound = min_val - buffer
            upper_bound = max_val + buffer

            return {
                param: (lower_bound[i].item(), upper_bound[i].item())
                for i, param in enumerate(param_names)
            }
        return None


class HistoryMatchingWorkflow(HistoryMatching):
    """
    History Matching Workflow class.

    Run history matching workflow:
    - sample parameter values to test from the current NROY parameter space
    - use emulator to rule out implausible parameters and update NROY space
    - make predictions for a subset the NROY parameters using the simulator
    - refit the emulator using the simulated data
    """

    def __init__(  # noqa: PLR0913 allow too many arguments since all currently required
        self,
        simulator: Simulator,
        emulator: ProbabilisticEmulator,
        observations: dict[str, tuple[float, float]] | dict[str, float],
        threshold: float = 3.0,
        model_discrepancy: float = 0.0,
        rank: int = 1,
        train_x: TensorLike | None = None,
        train_y: TensorLike | None = None,
        device: DeviceLike | None = None,
        random_seed: int | None = None,
    ):
        """
        Initialize the history matching workflow object.

        Parameters
        ----------
        simulator: Simulator
            A simulator.
        emulator: ProbabilisticEmulator
            A ProbabilisticEmulator pre-trained on `simulator` data.
        observations: dict[str, tuple[float, float] | dict[str, float]
            For each output variable, specifies observed [value, noise]. In case
            of no uncertainty in observations, provides just the observed value.
        threshold: float
            Implausibility threshold (query points with implausibility scores that
            exceed this value are ruled out). Defaults to 3, which is considered
            a good value for simulations with a single output.
        model_discrepancy: float
            Additional variance to include in the implausibility calculation.
        rank: int
            Scoring method for multi-output problems. Must be 1 <= rank <= n_outputs.
            When the implausibility scores are ordered across outputs, it indicates
            which rank to use when determining whether the query point is NROY. The
            default val of ``1`` indicates that the largest implausibility will be used.
        train_x: TensorLike | None
            Optional tensor of input data the emulator was trained on.
        train_y: TensorLike | None
            Optional tensor of output data the emulator was trained on.
        device: DeviceLike | None
            The device to use. If None, the default torch device is returned.
        """
        super().__init__(observations, threshold, model_discrepancy, rank, device)
        self.simulator = simulator
        if random_seed is not None:
            set_random_seed(seed=random_seed)
        self.emulator = emulator

        # These get updated when run() is called and used to refit the emulator
        if train_x is not None and train_y is not None:
            self.train_x = train_x.to(self.device)
            self.train_y = train_y.to(self.device)
        else:
            self.train_x = torch.empty((0, self.simulator.in_dim), device=self.device)
            self.train_y = torch.empty((0, self.simulator.out_dim), device=self.device)

    def generate_samples(self, n: int) -> tuple[TensorLike, TensorLike]:
        """
        Generate parameter samples and evaluate implausibility.

        Draw `n` samples from the simulator min/max parameter bounds and
        evaluate implausability given emulator predictions.

        Parameters
        ----------
        n: int
            The number of parameter samples to generate.

        Returns
        -------
        tuple[TensorLike, TensorLike]
            A tensor of tested input parameters and their implausability scores.
        """
        # Generate `n` parameter samples from within NROY bounds
        test_x = self.simulator.sample_inputs(n).to(self.device)

        # Rule out implausible parameters from samples using an emulator
        pred = self.emulator.predict(test_x)
        impl_scores = self.calculate_implausibility(pred.mean, pred.variance)

        return test_x, impl_scores

    def update_simulator_bounds(self, nroy_x: TensorLike, buffer_ratio: float = 0.05):
        """
        Update simulator parameter bounds to min/max of NROY parameter samples.

        Parameters
        ----------
        nroy_x: TensorLike
            A tensor of NROY parameter samples [n_samples, n_inputs]
        buffer_ratio: float
            A scaling factor used to expand the bounds of the (NROY) parameter space.
            It is applied as a ratio of the range (max_val - min_val) of each input
            parameter to create a buffer around the NROY minimum and maximum values.
        """
        # [n_outputs, 2] where second dim is [min, max]
        param_bounds = self.generate_param_bounds(nroy_x, buffer_ratio)
        if param_bounds is not None:
            self.simulator._param_bounds = list(param_bounds.values())
        else:
            warnings.warn(
                (
                    f"Could not update simulator parameter bounds only "
                    f"{nroy_x.shape[0]} samples were provided."
                ),
                stacklevel=2,
            )

    def sample_tensor(self, n: int, x: TensorLike) -> TensorLike:
        """
        Randomly sample `n` rows from `x`.

        Parameters
        ----------
        n: int
            The number of samples to draw.
        x: TensorLike
            The tensor to sample from.

        Returns
        -------
        TensorLike
            A tensor of samples with `n` rows.
        """
        if x.shape[0] < n:
            warnings.warn(
                f"Number of tensor rows {x.shape[0]} is less than {n} samples.",
                stacklevel=2,
            )
        idx = torch.randperm(x.shape[0], device=self.device)[:n]
        return x[idx]

    def simulate(self, x: TensorLike) -> tuple[TensorLike, TensorLike]:
        """
        Simulate `x` parameter inputs and filter out failed simulations.

        Parameters
        ----------
        x: TensorLike
            A tensor of parameters to simulate [n_samples, n_inputs].

        Returns
        -------
        tuple[TensorLike, TensorLike]
            Tensors of succesfully simulated input parameters and predictions.
        """
        # if simulation fails, returned y and x have fewer rows than input x
        y, x = self.simulator.forward_batch_skip_failures(x)
        y = y.to(self.device)
        x = x.to(self.device)

        self.train_y = torch.cat([self.train_y, y], dim=0)
        self.train_x = torch.cat([self.train_x, x], dim=0)

        return x, y

    def run(
        self,
        n_simulations: int = 100,
        n_test_samples: int = 10000,
        max_retries: int = 3,
        buffer_ratio: float = 0.0,
    ) -> tuple[TensorLike, TensorLike]:
        """
        Run a wave of the history matching workflow.

        Parameters
        ----------
        n_simulations: int
            The number of simulations to run.
        n_test_samples: int
            Number of input parameters to test for implausibility with the emulator.
            Parameters to simulate are sampled from this NROY subset.
        max_retries: int
            Maximum number of times to try to generate `n_simulations` NROY parameters.
            That is the maximum number of times to repeat the following steps:
                - draw `n_test_samples` parameters
                - use emulator to make predictions for those parameters
                - score implausability of parameters given predictions
                - identify NROY parameters within this set
        buffer_ratio: float
            A scaling factor used to expand the bounds of the (NROY) parameter space.
            It is applied as a ratio of the range (max_val - min_val) of each input
            parameter to create a buffer around the NROY minimum and maximum values
            when updating the simulator parameter bounds.

        Returns
        -------
        tuple[TensorLike, TensorLike]
            A tensor of tested input parameters and their implausibility scores from
            which simulation samples were then selected.
        """
        logger.debug(
            "Running history matching workflow with"
            " %d simulations and %d test samples.",
            n_simulations,
            n_test_samples,
        )

        test_parameters_list, impl_scores_list, nroy_parameters_list = (
            [],
            [],
            [torch.empty((0, self.simulator.in_dim), device=self.device)],
        )

        retries = 0
        while torch.cat(nroy_parameters_list, 0).shape[0] < n_simulations:
            if retries == max_retries:
                msg = (
                    f"Could not generate n_simulations ({n_simulations}) samples "
                    f"that are NROY after {max_retries} retries."
                )
                raise RuntimeError(msg)

            # Generate `n_test_samples` with implausability scores, identify NROY
            test_parameters, impl_scores = self.generate_samples(n_test_samples)
            nroy_parameters = self.get_nroy(impl_scores, test_parameters)

            # Store results
            nroy_parameters_list.append(nroy_parameters)
            test_parameters_list.append(test_parameters)
            impl_scores_list.append(impl_scores)

            retries += 1

        # Update simulator parameter bounds to NROY region
        # Next time that call run(), will sample from within this region
        nroy_parameters_all = torch.cat(nroy_parameters_list, 0)
        self.update_simulator_bounds(nroy_parameters_all, buffer_ratio)

        # Randomly pick at most `n_simulations` parameters from NROY to simulate
        nroy_simulation_samples = self.sample_tensor(n_simulations, nroy_parameters_all)

        # Make predictions using simulator (this updates self.x_train and self.y_train)
        _, _ = self.simulate(nroy_simulation_samples)

        # Refit emulator using all available data
        assert self.emulator is not None
        self.emulator.fit(self.train_x, self.train_y)

        # Return test parameters and impl scores for this run/wave
        return torch.cat(test_parameters_list, 0), torch.cat(impl_scores_list, 0)



from .pca import PCATransform
from .standardize import StandardizeTransform
from .vae import VAETransform

__all__ = ["PCATransform", "StandardizeTransform", "VAETransform"]

ALL_TRANSFORMS = [PCATransform, StandardizeTransform, VAETransform]

# Registry mapping transform names to their classes
# Add new transforms here when creating them
TRANSFORM_REGISTRY = {
    "pca": PCATransform,
    "standardize": StandardizeTransform,
    "vae": VAETransform,
}



from collections.abc import Sequence

import torch
from torch import Tensor

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import (
    DeviceLike,
    GaussianLike,
    TensorLike,
    TuneParams,
)
from autoemulate.emulators.base import (
    DropoutTorchBackend,
    Emulator,
    GaussianEmulator,
)
from autoemulate.emulators.nn.mlp import MLP
from autoemulate.transforms.standardize import StandardizeTransform
from autoemulate.transforms.utils import make_positive_definite


class Ensemble(GaussianEmulator):
    """
    Ensemble emulator that aggregates multiple Emulator instances to provide UQ.

    Ensemble emulator that aggregates multiple Emulator instances and returns
    a GaussianLike representing the ensemble posterior.
    Note that an Emulator instance may also be an Ensemble itself.
    """

    def __init__(
        self,
        emulators: Sequence[Emulator] | None = None,
        jitter: float = 1e-4,
        device: DeviceLike | None = None,
    ):
        """
        Initialize the Ensemble with a sequence of emulators.

        Parameters
        ----------
        emulators: Sequence[Emulator]
            A sequence of emulators to construct the ensemble with.
        jitter: float
            Amount of jitter to add to the covariance diagonal to avoid degeneracy.
            Defaults to 1e-4.
        device: DeviceLike | None
            The device to put torch tensors on.
        """
        assert isinstance(emulators, Sequence)
        for e in emulators:
            assert isinstance(e, Emulator)
        self.emulators = list(emulators)
        self.is_fitted_ = all(e.is_fitted_ for e in emulators)
        self.jitter = jitter
        self.supports_grad = all(e.supports_grad for e in self.emulators)
        TorchDeviceMixin.__init__(self, device=device)

    @staticmethod
    def is_multioutput() -> bool:
        """Ensemble supports multi-output."""
        return True

    @staticmethod
    def get_tune_params() -> TuneParams:
        """Return a dictionary of hyperparameters to tune."""
        return {}

    def _fit(self, x: TensorLike, y: TensorLike) -> None:
        for e in self.emulators:
            e.fit(x, y)
        self.is_fitted_ = True

    def _predict(self, x: Tensor, with_grad: bool) -> GaussianLike:
        if with_grad and not self.supports_grad:
            msg = "Gradient calculation is not supported."
            raise ValueError(msg)

        # Inference mode to disable autograd computation graph
        device = x.device
        means: list[Tensor] = []
        covs: list[Tensor] = []

        # Outputs from each emulator
        for e in self.emulators:
            out = e.predict(x, with_grad)
            if isinstance(out, GaussianLike):
                mu_i = out.mean.to(device)  # (batch_size, n_dims)
                assert isinstance(out.covariance_matrix, TensorLike)
                sigma_i = out.covariance_matrix.to(
                    device
                )  # (batch_size, n_dims, n_dims)
            elif isinstance(out, TensorLike):
                mu_i = out.to(device)
                # Instead of constructing dense zero matrix
                sigma_i = torch.broadcast_to(
                    torch.tensor(0.0),
                    (mu_i.shape[0], mu_i.shape[1], mu_i.shape[1]),
                ).to(device)
            else:
                s = f"Emulators of type {type(e)} are note supported yet."
                raise TypeError(s)
            means.append(mu_i)
            covs.append(sigma_i)

        # Stacked means and covs
        mu_stack = torch.stack(means, dim=0)  # (M, batch, dim)
        cov_stack = torch.stack(covs, dim=0)  # (M, batch, dim, dim)

        # Uniform ensemble mean and aleatoric average
        # NOTE: can implement weighting in future
        mu_ens = mu_stack.mean(dim=0)  # (batch, dim)
        sigma_alea = cov_stack.mean(dim=0)  # (batch, dim, dim)

        # Epistemic covariance: unbiased over M members
        dev = mu_stack - mu_ens.unsqueeze(0)  # (M, batch, dim)
        sigma_epi = torch.einsum("m b d, m b e -> b d e", dev, dev) / (
            len(self.emulators) - 1
        )

        # Total covariance
        sigma_ens = sigma_alea + sigma_epi  # (batch, dim, dim)

        # Return as MultivariateNormal
        return GaussianLike(
            mu_ens,
            make_positive_definite(
                sigma_ens, min_jitter=self.jitter, max_tries=3, clamp_eigvals=False
            ),
        )


class EnsembleMLP(Ensemble):
    """
    Ensemble of MLP emulators.

    This class is an ensemble of MLP emulators, each initialized with the same input and
    output data.

    """

    def __init__(  # noqa: PLR0913
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = True,
        standardize_y: bool = True,
        n_emulators: int = 4,
        device: DeviceLike | None = None,
        **mlp_kwargs,
    ):
        """
        Initialize an ensemble of MLPs.

        Parameters
        ----------
        x: TensorLike
            Input data tensor of shape (batch_size, n_features).
        y: TensorLike
            Target values tensor of shape (batch_size, n_outputs).
        standardize_x: bool
            Whether to standardize the input data. Defaults to True.
        standardize_y: bool
            Whether to standardize the output data. Defaults to True.
        n_emulators: int
            Number of MLP emulators to create in the ensemble. Defaults to 4.
        device: DeviceLike | None
            Device to run the model on (e.g., "cpu", "cuda"). Defaults to None.
        **mlp_kwargs: dict
            Additional keyword arguments for the MLP constructor.
        """
        emulators = [
            MLP(
                x,
                y,
                standardize_x=standardize_x,
                standardize_y=standardize_y,
                device=device,
                **mlp_kwargs,
            )
            for i in range(n_emulators)
        ]
        super().__init__(emulators, device=device)

    @staticmethod
    def is_multioutput() -> bool:
        """Ensemble of MLPs supports multi-output."""
        return True

    @staticmethod
    def get_tune_params() -> TuneParams:
        """Return a dictionary of hyperparameters to tune."""
        return {"n_emulators": [2, 4, 6, 8], **MLP.get_tune_params()}


class DropoutEnsemble(GaussianEmulator, TorchDeviceMixin):
    """
    Monte-Carlo Dropout ensemble.

    DropoutEnsemble does a number of forward passes with dropout on, and computes mean
    and epistemic covariance across them.
    """

    def __init__(  # noqa: PLR0913
        self,
        model: DropoutTorchBackend,
        standardize_x: bool = True,
        standardize_y: bool = True,
        n_samples: int = 20,
        jitter: float = 1e-4,
        device: DeviceLike | None = None,
    ):
        """
        Initialize the DropoutEnsemble with a fitted model.

        Parameters
        ----------
        model: DropoutTorchBackend
            A fitted DropoutTorchBackend (or any nn.Module with dropout layers).
        n_samples: int
            Number of forward passes to perform.
        jitter: float
            Amount of jitter to add to covariance diagonal to avoide degeneracy.
        device: DeviceLike | None
            torch device for inference (e.g. "cpu", "cuda").
        """
        assert isinstance(model, DropoutTorchBackend), "model must be a PyTorchBackend"
        TorchDeviceMixin.__init__(self, device=device)
        assert n_samples > 0
        self.model = model.to(self.device)
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.n_samples = n_samples
        self.is_fitted_ = model.is_fitted_
        self.jitter = jitter
        self.supports_grad = True

    @staticmethod
    def is_multioutput() -> bool:
        """DropoutEnsemble supports multi-output."""
        return True

    @staticmethod
    def get_tune_params() -> TuneParams:
        """Return a dictionary of hyperparameters to tune."""
        return {
            "n_samples": [10, 20, 50, 100],
        }

    def _fit(self, x: TensorLike, y: TensorLike) -> None:
        # Delegate training to the wrapped model
        self.model.fit(x, y)
        self.is_fitted_ = True

    def _predict(self, x: Tensor, with_grad: bool) -> GaussianLike:
        if not self.is_fitted_:
            s = "DropoutEnsemble: model is not fitted yet."
            raise RuntimeError(s)

        # move input to right device
        x = x.to(self.device)

        # enable dropout
        self.model.train()

        # collect M outputs
        samples = []
        with torch.set_grad_enabled(with_grad):
            for _ in range(self.n_samples):
                out = self.model.forward(x)
                # out: Tensor of shape (batch_size, output_dim)
                samples.append(out)

        # stack to shape (M, batch, dim)
        stack = torch.stack(samples, dim=0)

        # ensemble mean: (batch, dim)
        mu = stack.mean(dim=0)

        # compute epistemic covariance for each batch point
        # dev shape (M, batch, dim)
        dev = stack - mu.unsqueeze(0)
        # sigma_epi: (batch, dim, dim)
        sigma_epi = torch.einsum("m b d, m b e -> b d e", dev, dev) / (
            self.n_samples - 1
        )

        return GaussianLike(
            mu,
            make_positive_definite(
                sigma_epi, min_jitter=self.jitter, max_tries=3, clamp_eigvals=False
            ),
        )


class EnsembleMLPDropout(DropoutEnsemble):
    """Ensemble of MLP emulators with dropout."""

    def __init__(  # noqa: PLR0913
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = True,
        standardize_y: bool = True,
        dropout_prob: float = 0.2,
        device: DeviceLike | None = None,
        **mlp_kwargs,
    ):
        """
        Initialize an ensemble of MLPs with dropout.

        Parameters
        ----------
        x: TensorLike
            Input data tensor of shape (batch_size, n_features).
        y: TensorLike
            Target values tensor of shape (batch_size, n_outputs).
        standardize_x: bool
            Whether to standardize the input data. Defaults to True.
        standardize_y: bool
            Whether to standardize the output data. Defaults to True.
        dropout_prob: float
            Dropout probability to use in the MLP layers. Defaults to 0.2.
        device: DeviceLike | None
            Device to run the model on (e.g., "cpu", "cuda"). Defaults to None.
        **mlp_kwargs: dict
            Additional keyword arguments for the MLP constructor.
        """
        super().__init__(
            MLP(
                x,
                y,
                standardize_x=standardize_x,
                standardize_y=standardize_y,
                dropout_prob=dropout_prob,
                device=device,
                **mlp_kwargs,
            ),
            device=device,
        )

    @staticmethod
    def get_tune_params() -> TuneParams:
        """Return a dictionary of hyperparameters to tune."""
        params = MLP.get_tune_params()
        params["dropout_prob"] = [el for el in params["dropout_prob"] if el is not None]
        return {"n_emulators": [2, 4, 6, 8], **params}



import torch
from torch.distributions import Transform, constraints

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import TensorLike
from autoemulate.transforms.base import AutoEmulateTransform


class StandardizeTransform(AutoEmulateTransform):
    """
    Standardize transform for normalizing data.

    This transform is effectively a composition of two AffineTransforms with a
    translation by the mean and a scaling by the inverse of the standard deviation.

    The transform assumes that the input data is a matrix of shape `(n, d)`, where `n`
    is the number of samples and `d` is the number of features.
    """

    domain = constraints.real
    codomain = constraints.real
    bijective = True

    def __init__(self):
        # Cache not used as not expected to be beneficial for standardization
        Transform.__init__(self, cache_size=0)

    def fit(self, x: TensorLike):
        """Fit the StandardizeTransform to the data."""
        TorchDeviceMixin.__init__(self, device=x.device)

        self.check_matrix(x)
        self.mean = x.mean(0, keepdim=True)
        std = x.std(0, keepdim=True)
        # Ensure std is not zero to avoid division by zero errors
        std[std < 10 * torch.finfo(std.dtype).eps] = 1.0
        self.std = std
        self._is_fitted = True

    def _call(self, x):
        self._check_is_fitted()
        return (x - self.mean) / self.std

    def _inverse(self, y):
        self._check_is_fitted()
        return y * self.std + self.mean

    def log_abs_det_jacobian(self, x, y):
        """Compute the log absolute determinant of the Jacobian."""
        _, _ = x, y
        self._check_is_fitted()
        return torch.abs(self.std).log().sum()

    @property
    def _basis_matrix(self) -> TensorLike:
        return torch.eye(self.std.shape[1], device=self.device) * self.std



from typing import Literal

import numpy as np
from scipy.stats import loguniform
from sklearn.ensemble import GradientBoostingRegressor

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, TensorLike
from autoemulate.emulators.base import SklearnBackend
from autoemulate.transforms.standardize import StandardizeTransform


class GradientBoosting(SklearnBackend):
    """
    Gradient Boosting Emulator.

    Wraps Gradient Boosting regression from scikit-learn.
    """

    def __init__(  # noqa: PLR0913 allow too many arguments since all currently required
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = False,
        standardize_y: bool = False,
        loss: Literal[
            "squared_error", "absolute_error", "huber", "quantile"
        ] = "squared_error",
        learning_rate: float = 0.1,
        n_estimators: int = 100,
        max_depth: int = 3,
        min_samples_split: int = 2,
        min_samples_leaf: int = 1,
        subsample: float = 1.0,
        max_features: float | None = None,
        ccp_alpha: float = 0.0,
        n_iter_no_change: int | None = None,
        random_seed: int | None = None,
        device: DeviceLike = "cpu",
    ):
        """Initialize a GradientBoosting object.

        Parameters
        ----------
        x: TensorLike
            Input features.
        y: TensorLike
            Target values.
        standardize_x: bool
            Whether to standardize input features. Defaults to False.
        standardize_y: bool
            Whether to standardize target values. Defaults to False.
        loss: str
            Loss function to be optimized. Defaults to "squared_error".
        learning_rate: float
            Learning rate shrinks the contribution of each new tree. Defaults to 0.1.
        n_estimators: int
            The number of boosting stages to be run. Defaults to 100.
        max_depth: int
            Maximum depth of the individual regression estimators. Defaults to 3.
        min_samples_split: int
            Minimum number of samples required to split an internal node. Defaults to 2.
        min_samples_leaf: int
            Minimum number of samples required to be at a leaf node. Defaults to 1.
        subsample: float
            The fraction of samples to be used for fitting the individual base learners.
            Defaults to 1.0.
        max_features: float | None
            The number of features to consider when looking for the best split. Defaults
            to None.
        ccp_alpha: float
            Complexity parameter used for Minimal Cost-Complexity Pruning. Defaults to
            0.0.
        n_iter_no_change: int | None
            If not None, the number of iterations with no improvement to wait before
            stopping. Defaults to None.
        random_seed: int | None
            Random seed for reproducibility. If None, no seed is set. Defaults to None.
        device: DeviceLike
            Device to run the model on (e.g., "cpu", "cuda", "mps"). Defaults to "cpu".
        """
        _, _ = x, y  # ignore unused arguments
        TorchDeviceMixin.__init__(self, device=device, cpu_only=True)
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.loss = loss
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.subsample = subsample
        self.max_features = max_features
        self.ccp_alpha = ccp_alpha
        self.n_iter_no_change = n_iter_no_change
        self.random_seed = random_seed
        self.model = GradientBoostingRegressor(
            loss=self.loss,
            learning_rate=self.learning_rate,
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            subsample=self.subsample,
            max_features=self.max_features,
            ccp_alpha=self.ccp_alpha,
            n_iter_no_change=self.n_iter_no_change,
            random_state=self.random_seed,
        )

    @staticmethod
    def is_multioutput() -> bool:
        """GradientBoosting does not support multi-output."""
        return False

    @staticmethod
    def get_tune_params():
        """Return a dictionary of hyperparameters to tune."""
        return {
            "learning_rate": [loguniform(0.01, 0.2).rvs()],
            "n_estimators": [np.random.randint(100, 500)],
            "max_depth": [np.random.randint(3, 8)],
            "min_samples_split": [np.random.randint(2, 20)],
            "min_samples_leaf": [np.random.randint(1, 6)],
            "subsample": [np.random.uniform(0.6, 1.0)],
            "max_features": ["sqrt", "log2", None],
            "ccp_alpha": [loguniform(0.001, 0.1).rvs()],
        }



import random
from abc import ABC, abstractmethod
from typing import ClassVar

import numpy as np
import torch
from sklearn.base import BaseEstimator
from torch import nn, optim
from torch.distributions import TransformedDistribution
from torch.optim.lr_scheduler import ExponentialLR, LRScheduler

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import (
    DistributionLike,
    GaussianLike,
    NumpyLike,
    OutputLike,
    TensorLike,
    TuneParams,
)
from autoemulate.data.utils import ConversionMixin, ValidationMixin
from autoemulate.transforms.standardize import StandardizeTransform


class Emulator(ABC, ValidationMixin, ConversionMixin, TorchDeviceMixin):
    """
    Base class for all emulators.

    This class provides the basic structure and methods for emulators in AutoEmulate.
    It includes methods for fitting, predicting, and handling device management.

    """

    is_fitted_: bool = False
    supports_grad: bool = False
    scheduler_cls: type[optim.lr_scheduler.LRScheduler] | None = None
    x_transform: StandardizeTransform | None = None
    y_transform: StandardizeTransform | None = None

    @abstractmethod
    def _fit(self, x: TensorLike, y: TensorLike): ...

    def fit(self, x: TensorLike, y: TensorLike):
        """Fit the emulator to the provided data."""
        self._check(x, y)
        # Ensure x and y are tensors and 2D
        x, y = self._convert_to_tensors(x, y)

        # Move to device
        x, y = self._move_tensors_to_device(x, y)

        # Fit transforms
        if self.x_transform is not None:
            self.x_transform.fit(x)
        if self.y_transform is not None:
            self.y_transform.fit(y)
        x = self.x_transform(x) if self.x_transform is not None else x
        y = self.y_transform(y) if self.y_transform is not None else y

        # Fit emulator
        self._fit(x, y)
        self.is_fitted_ = True

    @abstractmethod
    def __init__(
        self, x: TensorLike | None = None, y: TensorLike | None = None, **kwargs
    ): ...

    @classmethod
    def model_name(cls) -> str:
        """Return the full name of the model."""
        return cls.__name__

    @classmethod
    def short_name(cls) -> str:
        """
        Return a short name for the model.

        Take the capital letters of the class name and return them as a lower case
        string. For example, if the class name is `GaussianProcess`, this will return
        `gp`.
        """
        return "".join([c for c in cls.__name__ if c.isupper()]).lower()

    @abstractmethod
    def _predict(self, x: TensorLike, with_grad: bool) -> OutputLike:
        pass

    def predict(self, x: TensorLike, with_grad: bool = False) -> OutputLike:
        """Predict the output for the given input.

        Parameters
        ----------
        x: TensorLike
            Input tensor to make predictions for.
        with_grad: bool
            Whether to enable gradient calculation. Defaults to False.

        Returns
        -------
        OutputLike
            The predicted output.
        """
        if not self.is_fitted_:
            msg = "Model is not fitted yet. Call fit() before predict()."
            raise RuntimeError(msg)
        self._check(x, None)
        (x,) = self._move_tensors_to_device(x)
        x = self.x_transform(x) if self.x_transform is not None else x
        output = self._predict(x, with_grad)
        if self.y_transform is not None:
            if isinstance(output, GaussianLike):
                output = self.y_transform._inverse_gaussian(output)
            elif isinstance(output, DistributionLike):
                output = TransformedDistribution(
                    output, transforms=[self.y_transform.inv]
                )
            elif isinstance(output, TensorLike):
                output = self.y_transform.inv(output)
            else:
                msg = (
                    "Output type not supported for transformation. "
                    f"Got {type(output)} but expected GaussianLike, "
                    "DistributionLike, or TensorLike."
                )
                raise TypeError(msg)
        self._check_output(output)
        return output

    @staticmethod
    @abstractmethod
    def is_multioutput() -> bool:
        """Flag to indicate if the model is multioutput or not."""

    @staticmethod
    def get_tune_params() -> TuneParams:
        """
        Return a dictionary of hyperparameters to tune.

        The keys in the TuneParams must be implemented as keyword arguments in the
        __init__ method of any subclasses.

        e.g.

        tune_params: TuneParams = {
            "lr": list[0.01, 0.1],
            "batch_size": [16, 32],
            "mean"
        }

        model_params: ModelParams = {
            "lr": 0.01,
            "batch_size": 16
        }

        class MySubClass(Emulator):
            def __init__(lr, batch_size):
                self.lr = lr
                self.batch_size = batch_size
        """
        msg = (
            "Subclasses should implement for generating tuning params specific to "
            "each subclass."
        )
        raise NotImplementedError(msg)

    @classmethod
    def get_random_params(cls):
        """Return a random set of params for the model."""
        return {
            k: v[np.random.randint(len(v))] for k, v in cls.get_tune_params().items()
        }

    @classmethod
    def scheduler_params(cls) -> dict:
        """
        Return a random parameters for the learning rate scheduler.

        This should be added to the `get_tune_params()` method of subclasses
        to allow tuning of the scheduler parameters.
        """
        all_params = [
            {
                "scheduler_cls": [None],
                "scheduler_kwargs": [{}],
            },
            {
                "scheduler_cls": [ExponentialLR],
                "scheduler_kwargs": [
                    {"gamma": 0.9},
                    {"gamma": 0.95},
                ],
            },
            {
                "scheduler_cls": [LRScheduler],
                "scheduler_kwargs": [
                    {"policy": "ReduceLROnPlateau", "patience": 5, "factor": 0.5}
                ],
            },
            # TODO: investigate these suggestions from copilot, issue: #597
            # {
            #     "scheduler_cls": [CosineAnnealingLR],
            #     "scheduler_kwargs": [{"T_max": 10, "eta_min": 0.01}],
            # },
            # {
            #     "scheduler_cls": [ReduceLROnPlateau],
            #     "scheduler_kwargs": [{"mode": "min", "factor": 0.1, "patience": 5}],
            # },
            # {
            #     "scheduler_cls": [StepLR],
            #     "scheduler_kwargs": [{"step_size": 10, "gamma": 0.1}],
            # },
            # {
            #     "scheduler_cls": [CyclicLR],
            #     "scheduler_kwargs": [{
            #         "base_lr": 1e-3,
            #         "max_lr": 1e-1,
            #         "step_size_up": 5,
            #         "step_size_down": 5,
            #     }],
            # },
            # {
            #     "scheduler_cls": [OneCycleLR],
            #     "scheduler_kwargs": [{
            #         "max_lr": 1e-1,
            #         "total_steps": self.epochs,
            #         "pct_start": 0.3,
            #         "anneal_strategy": "linear",
            #     }],
            # },
        ]
        # Randomly select one of the parameter sets
        return random.choice(all_params)

    def scheduler_setup(self, kwargs: dict | None = None):
        """
        Set up the learning rate scheduler for the emulator.

        Parameters
        ----------
        kwargs : dict | None
            Keyword arguments for the model. This should include scheduler_kwargs.
        """
        if kwargs is None:
            msg = (
                "Provide a kwargs dictionary including "
                "scheduler_kwargs to set up the scheduler."
            )
            raise ValueError(msg)

        if not hasattr(self, "optimizer"):
            msg = "Optimizer must be set before setting up the scheduler."
            raise RuntimeError(msg)

        # Extract scheduler-specific kwargs if present
        try:
            assert type(kwargs) is dict
            scheduler_kwargs = kwargs.pop("scheduler_kwargs", {})
        except AttributeError:
            # If kwargs does not contain scheduler_kwargs, throw an error
            msg = "No kwargs for scheduler setup detected."
            raise ValueError(msg) from None

        # Set up the scheduler if a scheduler class is defined
        if self.scheduler_cls is None:
            self.scheduler = None
        else:
            self.scheduler = self.scheduler_cls(self.optimizer, **scheduler_kwargs)  # type: ignore[call-arg]


class DeterministicEmulator(Emulator):
    """A base class for deterministic emulators."""

    @abstractmethod
    def _predict(self, x: TensorLike, with_grad: bool) -> TensorLike: ...
    def predict(self, x: TensorLike, with_grad: bool = False) -> TensorLike:
        """Predict the output for the given input.

        Parameters
        ----------
        x: TensorLike
            Input tensor to make predictions for.
        with_grad: bool
            Whether to enable gradient calculation. Defaults to False.

        Returns
        -------
        TensorLike
            The emulator predicted output for `x`.
        """
        pred = super().predict(x, with_grad)
        assert isinstance(pred, TensorLike)
        return pred


class ProbabilisticEmulator(Emulator):
    """A base class for probabilistic emulators."""

    @abstractmethod
    def _predict(self, x: TensorLike, with_grad: bool) -> DistributionLike: ...
    def predict(self, x: TensorLike, with_grad: bool = False) -> DistributionLike:
        """Predict the output distribution for the given input.

        Parameters
        ----------
        x: TensorLike
            Input tensor to make predictions for.
        with_grad: bool
            Whether to enable gradient calculation. Defaults to False.

        Returns
        -------
        DistributionLike
            The emulator predicted distribution for `x`.
        """
        pred = super().predict(x, with_grad)
        assert isinstance(pred, DistributionLike)
        return pred

    def predict_mean_and_variance(
        self, x: TensorLike, with_grad: bool = False
    ) -> tuple[TensorLike, TensorLike]:
        """
        Predict mean and variance from the probabilistic output.

        Parameters
        ----------
        x: TensorLike
            Input tensor to make predictions for.
        with_grad: bool
            Whether to enable gradient calculation. Defaults to False.

        Returns
        -------
        tuple[TensorLike, TensorLike]
            The emulator predicted mean and variance for `x`.
        """
        pred = self.predict(x, with_grad)
        return pred.mean, pred.variance


class GaussianEmulator(ProbabilisticEmulator):
    """A base class for Gaussian emulators."""

    supports_grad: bool = True

    @abstractmethod
    def _predict(self, x: TensorLike, with_grad: bool) -> GaussianLike: ...
    def predict(self, x: TensorLike, with_grad: bool = False) -> GaussianLike:
        """Predict the Gaussian distribution for the given input.

        Parameters
        ----------
        x: TensorLike
            Input tensor to make predictions for.
        with_grad: bool
            Whether to enable gradient calculation. Defaults to False.

        Returns
        -------
        GaussianLike
            The emulator predicted Gaussian distribution for `x`.
        """
        pred = super().predict(x, with_grad)
        assert isinstance(pred, GaussianLike)
        return pred


class GaussianProcessEmulator(GaussianEmulator):
    """A base class for Gaussian Process emulators."""

    @abstractmethod
    def _predict(self, x: TensorLike, with_grad: bool) -> GaussianLike: ...
    def predict(self, x: TensorLike, with_grad: bool = False) -> GaussianLike:
        """Predict the Gaussian distribution for the given input.

        Parameters
        ----------
        x: TensorLike
            Input tensor to make predictions for.
        with_grad: bool
            Whether to enable gradient calculation. Defaults to False.

        Returns
        -------
        GaussianLike
            The emulator predicted Gaussian distribution for `x`.
        """
        pred = super().predict(x, with_grad)
        assert isinstance(pred, GaussianLike)
        return pred


class PyTorchBackend(nn.Module, Emulator):
    """
    `PyTorchBackend` provides a backend for PyTorch models.

    The class provides the basic structure and methods for PyTorch-based emulators to
    enable further subclassing and customization. This provides default implementations
    to simplify model-specific subclasses by only needing to implement:
      - `.__init__()`: the constructor for the model
      - `.forward()`: the forward pass of the model
      - `.get_tune_params()`: the hyperparameters to tune for the model

    """

    batch_size: int = 16
    shuffle: bool = True
    epochs: int = 10
    loss_history: ClassVar[list[float]] = []
    verbose: bool = False
    loss_fn: nn.Module = nn.MSELoss()
    optimizer_cls: type[optim.Optimizer] = optim.Adam
    optimizer: optim.Optimizer
    supports_grad: bool = True
    lr: float = 1e-1
    scheduler_cls: type[LRScheduler] | None = None

    def loss_func(self, y_pred, y_true):
        """Loss function to be used for training the model."""
        return nn.MSELoss()(y_pred, y_true)

    def _fit(
        self,
        x: TensorLike,
        y: TensorLike,
    ):
        """
        Train a PyTorchBackend model.

        Parameters
        ----------
        x: TensorLike
            Input features as numpy array, PyTorch tensor, or DataLoader.
        y: OutputLike or None
            Target values (not needed if x is a DataLoader).
        """
        self.train()  # Set model to training mode

        # Convert input to DataLoader if not already
        dataloader = self._convert_to_dataloader(
            x, y, batch_size=self.batch_size, shuffle=self.shuffle
        )

        # Training loop
        for epoch in range(self.epochs):
            epoch_loss = 0.0
            batches = 0

            for x_batch, y_batch in dataloader:
                # Forward pass
                y_pred = self.forward(x_batch)
                loss = self.loss_func(y_pred, y_batch)

                # Backward pass and optimize
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                # Track loss
                epoch_loss += loss.item()
                batches += 1
            # Update learning rate if scheduler is defined
            if self.scheduler is not None:
                self.scheduler.step()  # type: ignore[call-arg]

            # Average loss for the epoch
            avg_epoch_loss = epoch_loss / batches
            self.loss_history.append(avg_epoch_loss)

            if self.verbose and (epoch + 1) % (self.epochs // 10 or 1) == 0:
                print(f"Epoch [{epoch + 1}/{self.epochs}], Loss: {avg_epoch_loss:.4f}")

    def _initialize_weights(
        self,
        weight_init: str = "default",
        scale: float = 1.0,
        bias_init: str = "default",
    ):
        """
        Initialize the weights.

        Parameters
        ----------
        weight_init: str
            Initialization method name
        scale: float
            Scale parameter for initialization methods. Used as:
            - gain for Xavier methods
            - std for normal distribution
            - bound for uniform distribution (range: [-scale, scale])
            - ignored for Kaiming methods (uses optimal scaling)
        bias_init: str
            Bias initialization method. Options: "zeros", "default":
                - "zeros" initializes biases to zero
                - "default" uses PyTorch's default uniform initialization
        """
        # Dictionary mapping for weight initialization methods
        init_methods = {
            "xavier_uniform": lambda w: nn.init.xavier_uniform_(w, gain=scale),
            "xavier_normal": lambda w: nn.init.xavier_normal_(w, gain=scale),
            "kaiming_uniform": lambda w: nn.init.kaiming_uniform_(
                w, mode="fan_in", nonlinearity="relu"
            ),
            "kaiming_normal": lambda w: nn.init.kaiming_normal_(
                w, mode="fan_in", nonlinearity="relu"
            ),
            "normal": lambda w: nn.init.normal_(w, mean=0.0, std=scale),
            "uniform": lambda w: nn.init.uniform_(w, -scale, scale),
            "zeros": lambda w: nn.init.zeros_(w),
            "ones": lambda w: nn.init.ones_(w),
        }

        for module in self.modules():
            # TODO: consider and add handling for other module types
            if isinstance(module, nn.Linear):
                # Apply initialization if method exists
                if weight_init in init_methods:
                    init_methods[weight_init](module.weight)

                # Initialize biases based on bias_init parameter
                if module.bias is not None and bias_init == "zeros":
                    nn.init.zeros_(module.bias)

    def _predict(self, x: TensorLike, with_grad: bool) -> OutputLike:
        self.eval()
        with torch.set_grad_enabled(with_grad):
            return self(x)


class SklearnBackend(DeterministicEmulator):
    """
    `SklearnBackend` provides a backend for sklearn models.

    The class provides the basic structure and methods for sklearn-based emulators to
    enable further subclassing and customization. This provides default implementations
    to simplify model-specific subclasses by only needing to implement:
      - `.__init__()`: the constructor for the model
      - `.get_tune_params()`: the hyperparameters to tune for the model
    """

    model: BaseEstimator
    normalize_y: bool = False
    y_mean: TensorLike
    y_std: TensorLike
    supports_grad: bool = False

    def _model_specific_check(self, x: NumpyLike, y: NumpyLike):
        _, _ = x, y

    def _fit(self, x: TensorLike, y: TensorLike):
        if self.normalize_y:
            y, y_mean, y_std = self._normalize(y)
            self.y_mean = y_mean
            self.y_std = y_std
        x_np, y_np = self._convert_to_numpy(x, y)
        assert isinstance(x_np, np.ndarray)
        assert isinstance(y_np, np.ndarray)
        self.n_features_in_ = x_np.shape[1]
        self._model_specific_check(x_np, y_np)
        self.model.fit(x_np, y_np)  # type: ignore PGH003

    def _predict(self, x: TensorLike, with_grad: bool) -> TensorLike:
        if with_grad:
            msg = "Gradient calculation is not supported."
            raise ValueError(msg)
        x_np, _ = self._convert_to_numpy(x, None)
        y_pred = self.model.predict(x_np)  # type: ignore PGH003
        _, y_pred = self._move_tensors_to_device(*self._convert_to_tensors(x, y_pred))
        if self.normalize_y:
            y_pred = self._denormalize(y_pred, self.y_mean, self.y_std)
        return y_pred


class DropoutTorchBackend(PyTorchBackend):
    """PyTorch backend model that is able to support dropout."""



# import functools

# import numpy as np
# import torch
# import torch.utils
# import torch.utils.data
# from torch import nn
# from torch.utils.data import Dataset

# from autoemulate.device import TorchDeviceMixin, get_torch_device
# from autoemulate.emulators.base import PyTorchBackend
# from autoemulate.types import DeviceLike, DistributionLike, TensorLike


# class CNPDataset(Dataset, TorchDeviceMixin):
#     """
#     Dataset for Conditional Neural Process (CNP).
#     Samples context points and target points for
#     each episode.
#     """

#     def __init__(
#         self,
#         x: TensorLike,
#         y: TensorLike,
#         min_context_points: int,
#         max_context_points: int,
#         n_episode: int,
#         device: DeviceLike | None = None,
#     ):
#         """
#         Parameters
#         ----------
#         x: (n_samples, x_dim)
#             Input data
#         y: (n_samples, y_dim)
#             Output data
#         min_context_points: int
#             Minimum number of context points to sample
#         max_context_points: int
#             Maximum number of context points to sample
#         n_episode: int
#             Number of episodes to sample. Must be greater than max_context_points.
#         """
#         TorchDeviceMixin.__init__(self, device=device)
#         self.x, self.y = self._move_tensors_to_device(x, y)
#         if max_context_points >= n_episode:
#             msg = "max_context_points must be less than n_episode"
#             raise ValueError(msg)
#         self.max_context_points = max_context_points
#         self.min_context_points = min_context_points
#         self.n_samples = len(self.x)
#         # self.x_dim = self.x[0][0].shape[-1]
#         self.n_episode = n_episode

#     def __len__(self):
#         return len(self.x)

#     def __getitem__(self, idx):
#         return self.sample()

#     def sample(self):
#         """
#         From the full dataset, sample context and target points.
#         """

#         x = self.x
#         y = self.y

#         # sample context points
#         num_context = np.random.randint(
#             self.min_context_points,
#             self.max_context_points + 1,
#         )

#         # randomly select context
#         perm = torch.randperm(self.n_samples)
#         context_idxs = perm[:num_context]
#         target_idxs = perm[
#             : self.n_episode
#         ]  # including context points in targets helps with training

#         x_context = x[context_idxs]
#         y_context = y[context_idxs]
#         x_target = x[target_idxs]
#         y_target = y[target_idxs]

#         x = {
#             "x_context": x_context,
#             "y_context": y_context,
#             "x_target": x_target,
#         }

#         return (x, y_target)


# def cnp_collate_fn(batch, device: DeviceLike | None = None):
#     """
#     Collate function for CNP.

#     Handles different dimensions in each sample due to different numbers of context
#     points.

#     TODO:
#     this currently just adds zeros. These shouldn't have an impact on the output as long # noqa: E501
#     as we don't do layernorm or attention, but we should modify cnp etc. to handle this # noqa: E501
#     better.
#     """
#     X, y = zip(*batch, strict=False)
#     device = get_torch_device(device)

#     # Get the maximum number of context and target points in the batch
#     max_context = max(x["x_context"].shape[0] for x in X)
#     max_target = max(x["x_target"].shape[0] for x in X)

#     # Initialize tensors to hold the batched data
#     x_context_batched = torch.zeros(len(batch), max_context, X[0]["x_context"].shape[1]) # noqa: E501
#     y_context_batched = torch.zeros(len(batch), max_context, X[0]["y_context"].shape[1]) # noqa: E501
#     x_target_batched = torch.zeros(len(batch), max_target, X[0]["x_target"].shape[1])
#     y_batched = torch.zeros(len(batch), max_target, y[0].shape[1])

#     # Create masks for context and target
#     context_mask = torch.zeros(len(batch), max_context, dtype=torch.bool)
#     target_mask = torch.zeros(len(batch), max_target, dtype=torch.bool)

#     # Fill in the batched tensors
#     for i, (x, yi) in enumerate(zip(X, y, strict=False)):
#         n_context = x["x_context"].shape[0]
#         n_target = x["x_target"].shape[0]

#         x_context_batched[i, :n_context] = x["x_context"]
#         y_context_batched[i, :n_context] = x["y_context"]
#         x_target_batched[i, :n_target] = x["x_target"]
#         y_batched[i, :n_target] = yi

#         context_mask[i, :n_context] = True
#         target_mask[i, :n_target] = True

#     # Create a dictionary for X
#     x_batched = {
#         "x_context": x_context_batched.to(device),
#         "y_context": y_context_batched.to(device),
#         "x_target": x_target_batched.to(device),
#         "context_mask": context_mask.to(device),
#     }

#     return x_batched, y_batched.to(device)


# class Encoder(nn.Module):
#     """
#     Deterministic encoder for conditional neural process model.
#     """

#     def __init__(
#         self,
#         input_dim: int,
#         output_dim: int,
#         hidden_dim: int,
#         latent_dim: int,
#         hidden_layers_enc: int,
#         activation: type[nn.Module],
#     ):
#         super().__init__()
#         layers = [nn.Linear(input_dim + output_dim, hidden_dim), activation()]
#         for _ in range(hidden_layers_enc):
#             layers.extend([nn.Linear(hidden_dim, hidden_dim), activation()])
#         layers.append(nn.Linear(hidden_dim, latent_dim))
#         self.net = nn.Sequential(*layers)

#     def forward(
#         self,
#         x_context: TensorLike,
#         y_context: TensorLike,
#         context_mask: TensorLike | None = None,
#     ):
#         """
#         Encode context

#         Parameters
#         ----------
#         x_context: (batch_size, n_context_points, input_dim)
#         y_context: (batch_size, n_context_points, output_dim)
#         context_mask: (batch_size, n_context_points)

#         Returns
#         -------
#         r: (batch_size, 1, latent_dim)
#         """
#         x = torch.cat([x_context, y_context], dim=-1)
#         x = self.net(x)

#         if context_mask is not None:
#             masked_x = x * context_mask.unsqueeze(-1)
#             r = masked_x.sum(dim=1, keepdim=True) / context_mask.sum(
#                 dim=1, keepdim=True
#             ).unsqueeze(
#                 -1
#             )  # mean over valid context points
#         else:
#             r = x.mean(dim=1, keepdim=True)  # mean over context points
#         return r


# class Decoder(nn.Module):
#     def __init__(
#         self,
#         input_dim: int,
#         latent_dim: int,
#         hidden_dim: int,
#         output_dim: int,
#         hidden_layers_dec: int,
#         activation: type[nn.Module],
#     ):
#         super().__init__()
#         layers = [nn.Linear(latent_dim + input_dim, hidden_dim), activation()]
#         for _ in range(hidden_layers_dec):
#             layers.extend([nn.Linear(hidden_dim, hidden_dim), activation()])
#         self.net = nn.Sequential(*layers)
#         self.mean_head = nn.Linear(hidden_dim, output_dim)
#         self.logvar_head = nn.Linear(hidden_dim, output_dim)

#     def forward(self, r: TensorLike, x_target: TensorLike):
#         """
#         Decode using representation r and target points x_target

#         Parameters
#         ----------
#         r: (batch_size, 1, latent_dim)
#         x_target: (batch_size, n_points, input_dim)

#         Returns
#         -------
#         mean: (batch_size, n_points, output_dim)
#         logvar: (batch_size, n_points, output_dim)
#         """
#         _, n, _ = x_target.shape  # batch_size, n_points, input_dim
#         r_expanded = r.expand(-1, n, -1)
#         x = torch.cat([r_expanded, x_target], dim=-1)
#         hidden = self.net(x)
#         mean = self.mean_head(hidden)
#         logvar = self.logvar_head(hidden)

#         return mean, logvar


# class CNPModule(PyTorchBackend):
#     """ "
#     Implemntation of Conditional Neural Process (CNP) model.
#     The model is a deterministic encoder-decoder architecture.
#     """

#     def __init__(
#         self,
#         x: TensorLike,
#         y: TensorLike,
#         hidden_dim: int = 32,
#         latent_dim: int = 16,
#         hidden_layers_enc: int = 2,
#         hidden_layers_dec: int = 2,
#         activation: type[nn.Module] = nn.ReLU,
#         min_context_points: int = 2,
#         offset_context_points: int = 2,
#         n_episodes: int = 12,
#         batch_size: int = 4,
#         device: DeviceLike | None = None,
#     ):
#         """
#         Parameters
#         ----------
#         x: (n_points, input_dim)
#             Input data of shape (n_points, input_dim).
#         y: (n_points, output_dim)
#             Output data of shape (n_points, output_dim).
#         hidden_dim: int
#             Hidden dimension of the encoder and decoder.
#         latent_dim: int
#             Latent dimension of the encoder and decoder.
#         hidden_layers_enc: int
#             Number of hidden layers in the encoder.
#         hidden_layers_dec: int
#             Number of hidden layers in the decoder.
#         activation: type[nn.Module]
#             Activation function to use in the encoder and decoder.
#         min_context_points: int
#             Minimum number of context points to sample.
#         offset_context_points: int
#             Offset for the maximum number of context points to sample.
#             max_context_points = min_context_points + offset_context_points
#         n_episodes: int
#             This is the length of the target sequence.
#             It must be greater than max_context_points.
#         batch_size: int
#             Batch size for training.
#         device: DeviceLike | None
#             Device to use for training. If None, use the default device.
#         """
#         super().__init__()
#         TorchDeviceMixin.__init__(self, device=device)
#         x, y = self._move_tensors_to_device(x, y)

#         x, y = self._convert_to_tensors(x, y)
#         self.input_dim = x.shape[1]
#         self.output_dim = y.shape[1]

#         self.encoder = Encoder(
#             self.input_dim,
#             self.output_dim,
#             hidden_dim,
#             latent_dim,
#             hidden_layers_enc,
#             activation,
#         )
#         self.decoder = Decoder(
#             self.input_dim,
#             latent_dim,
#             hidden_dim,
#             self.output_dim,
#             hidden_layers_dec,
#             activation,
#         )

#         # Move to device
#         self.encoder.to(self.device)
#         self.decoder.to(self.device)
#         self.to(self.device)

#         self.min_context_points = min_context_points
#         self.max_context_points = self.min_context_points + offset_context_points
#         self.n_episode = n_episodes
#         self.optimizer = torch.optim.Adam(self.parameters(), lr=0.001)

#         self.x_train = None
#         self.y_train = None
#         self.batch_size = batch_size

#     def forward(
#         self,
#         context_x: TensorLike,
#         context_y: TensorLike,
#         target_x: TensorLike,
#         context_mask: TensorLike | None = None,
#     ) -> torch.distributions.Independent:
#         """

#         Parameters
#         ----------
#         context_x: (batch_size, n_context_points, input_dim)
#         context_y: (batch_size, n_context_points, output_dim)
#         target_x: (batch_size, n_points, input_dim)
#         context_mask: (batch_size, n_context_points), currently unused,
#         as we pad with 0's and don't have attention, layernorm yet.

#         Returns
#         -------
#         mean: (batch_size, n_points, output_dim)
#         logvar: (batch_size, n_points, output_dim)
#         """

#         r = self.encoder(context_x, context_y, context_mask)
#         mean, logvar = self.decoder(r, target_x)

#         return torch.distributions.Independent(
#             torch.distributions.Normal(mean, torch.exp(0.5 * logvar)),
#             reinterpreted_batch_ndims=1,
#         )

#     def _fit(self, x: TensorLike, y: TensorLike):
#         """
#         Fit the model to the data.
#         Note the batching of data is done internally in the method.
#         Parameters
#         ----------
#         x: (n_points, input_dim)
#             Input data of shape (n_points, input_dim).
#         y: (n_points, output_dim)
#             Output data of shape (n_points, output_dim).
#         """
#         self.train()
#         x, y = self._move_tensors_to_device(x, y)

#         # Save off all X_train and y_train
#         self.x_train, self.y_train = self._convert_to_tensors(x, y)
#         self.x_train, self.y_train = self._move_tensors_to_device(
#             self.x_train, self.y_train
#         )

#         # Convert dataset to CNP Dataset
#         dataset = CNPDataset(
#             self.x_train,
#             self.y_train,
#             min_context_points=self.min_context_points,
#             max_context_points=self.max_context_points,
#             n_episode=self.n_episode,
#         )

#         # Training loop
#         for epoch in range(self.epochs):
#             epoch_loss = 0.0
#             batches = 0

#             for x_batch, y_target in torch.utils.data.DataLoader(
#                 dataset,
#                 batch_size=self.batch_size,
#                 collate_fn=functools.partial(cnp_collate_fn, device=self.device),
#             ):
#                 x_context = x_batch["x_context"]
#                 y_context = x_batch["y_context"]
#                 x_target = x_batch["x_target"]

#                 # Preprocess x_batch
#                 x_context = self.preprocess(x_context)
#                 x_target = self.preprocess(x_target)
#                 assert isinstance(x_context, TensorLike)
#                 assert isinstance(x_target, TensorLike)

#                 # Forward pass
#                 y_pred = self.forward(x_context, y_context, x_target)
#                 loss = -y_pred.log_prob(y_target).sum(1).mean()

#                 # Backward pass and optimize
#                 self.optimizer.zero_grad()
#                 loss.backward()
#                 self.optimizer.step()

#                 # Track loss
#                 epoch_loss += loss.item()
#                 batches += 1

#             # Average loss for the epoch
#             avg_epoch_loss = epoch_loss / batches
#             self.loss_history.append(avg_epoch_loss)

#             if self.verbose and (epoch + 1) % (self.epochs // 10 or 1) == 0:
#                 print(f"Epoch [{epoch + 1}/{self.epochs}], Loss: {avg_epoch_loss:.4f}") # noqa: E501

#     def _predict(self, x: TensorLike) -> DistributionLike:
#         """
#         Predict uses the training data as the context data and the input x as the target # noqa: E501
#         data. The data is preprocessed within the method.

#         Parameters
#         ----------
#         x: (n_points, input_dim)
#             Input data of shape (n_points, input_dim).

#         Returns
#         -------
#         distribution: torch.distributions.Independent
#             Distribution of the output data.
#             Note the distribution is a single tensor of shape (n_points, output_dim).

#         """

#         self.eval()
#         x = x.to(self.device)
#         x = self.preprocess(x)

#         x_target = self._convert_to_tensors(x)

#         # Sort splitting into context and target
#         # TODO: Do we need a conversion for single tensors?
#         if not isinstance(x_target, TensorLike):
#             msg = "x_target must be a single tensor"
#             raise ValueError(msg)

#         # Unsqueeze the batch dimension for x_train, y_train and x_target
#         assert isinstance(self.x_train, TensorLike)
#         assert isinstance(self.y_train, TensorLike)
#         x_train = self.x_train.unsqueeze(0)
#         y_train = self.y_train.unsqueeze(0)
#         x_target = x_target.unsqueeze(0)

#         # Forward pass and drop the batch dimension
#         distribution = self.forward(x_train, y_train, x_target)
#         mean = distribution.mean.squeeze(0)  # Remove batch dimension
#         variance = distribution.variance.squeeze(0)  # Remove batch dimension
#         return torch.distributions.Independent(
#             torch.distributions.Normal(mean, variance.sqrt()),
#             reinterpreted_batch_ndims=1,
#         )

#     @staticmethod
#     def is_multioutput() -> bool:
#         """
#         Check if the model is a multi-output model.
#         """
#         return True

#     @staticmethod
#     def get_tune_config():
#         return {
#             "hidden_dim": [16, 32, 64],
#             "latent_dim": np.arange(8, 64, 8),
#             "hidden_layers_enc": [1, 2, 4],
#             "hidden_layers_dec": [1, 2, 4],
#             "activation": [nn.ReLU],
#             "min_context_points": [4, 5, 6],
#             "offset_context_points": [4, 5],
#             # max_context_points must be less than n_episodes
#             "n_episodes": [12, 13, 14],
#         }



import json
import logging

import numpy as np
from sklearn.model_selection import KFold
from torchmetrics import R2Score

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.model_selection import cross_validate
from autoemulate.core.types import DeviceLike, InputLike, ModelParams
from autoemulate.data.utils import set_random_seed
from autoemulate.emulators.base import ConversionMixin, Emulator
from autoemulate.transforms.base import AutoEmulateTransform

logger = logging.getLogger("autoemulate")


class Tuner(ConversionMixin, TorchDeviceMixin):
    """
    Run randomised hyperparameter search for a given model.

    Parameters
    ----------
    x: InputLike
        Input features.
    y: OutputLike or None
        Target values (not needed if x is a Dataset).
    n_iter: int
        Number of parameter settings to randomly sample and test.
    device: DeviceLike | None
        The device to use for model training and evaluation. If None, uses the default
        device (usually CPU or GPU).
    random_seed: int | None
        Random seed for reproducibility. If None, no seed is set.
    """

    def __init__(
        self,
        x: InputLike,
        y: InputLike | None,
        n_iter: int = 10,
        device: DeviceLike | None = None,
        random_seed: int | None = None,
    ):
        TorchDeviceMixin.__init__(self, device=device)
        self.n_iter = n_iter

        # Convert input types, convert to tensors to ensure correct shapes, move to
        # device and convert back to dataset. TODO: consider if this is the best way to
        # do this.
        dataset = self._convert_to_dataset(x, y)
        x_tensor, y_tensor = self._convert_to_tensors(dataset)
        x_tensor, y_tensor = self._move_tensors_to_device(x_tensor, y_tensor)
        self.dataset = self._convert_to_dataset(x_tensor, y_tensor)

        # Q: should users be able to choose a different validation metric?
        self.metric = R2Score

        if random_seed is not None:
            set_random_seed(seed=random_seed)

    def run(  # noqa: PLR0913
        self,
        model_class: type[Emulator],
        x_transforms: list[AutoEmulateTransform] | None = None,
        y_transforms: list[AutoEmulateTransform] | None = None,
        n_splits: int = 5,
        seed: int | None = None,
        shuffle: bool = True,
    ) -> tuple[list[list[float]], list[ModelParams]]:
        """
        Run randomised hyperparameter search for a given model.

        Parameters
        ----------
        model_class: type[Emulator]
            A concrete Emulator subclass.
        x_transforms_list: list[list[AutoEmulateTransform]] | None
            An optional list of sequences of transforms to apply to the input data.
            Defaults to None, in which case the data is standardized.
        y_transforms_list: list[list[AutoEmulateTransform]] | None
            An optional list of sequences of transforms to apply to the output data.
            Defaults to None, in which case the data is standardized.
        n_splits: int
            Number of cross validation folds to split data into. Defaults to 5.
        seed: int | None
            Random seed to use in cross validation. If None, no seed is set.
        shuffle: bool
            Whether to shuffle data before splitting into cross validation folds.
            Defaults to True.

        Returns
        -------
        Tuple[list[float], list[ModelParams]]
            The validation scores and parameter values used in each search iteration.
        """
        # keep track of what parameter values tested and how they performed
        model_params_tested: list[ModelParams] = []
        val_scores: list[list[float]] = []

        # Initialize retries and maximum number of retries for consecutive failed tuning
        # iterations that would indicate that no model params is working for the
        # given dataset/model and that the tuning process should be stopped
        retries, max_retries = 0, 100
        while len(model_params_tested) < self.n_iter:
            # randomly sample hyperparameters and instantiate model
            model_params = model_class.get_random_params()
            try:
                # Perform cross-validation on randomly sampled model params
                scores = cross_validate(
                    cv=KFold(n_splits=n_splits, random_state=seed, shuffle=shuffle),
                    dataset=self.dataset,
                    x_transforms=x_transforms,
                    y_transforms=y_transforms,
                    model=model_class,
                    model_params=model_params,
                    device=self.device,
                    random_seed=None,
                )

                # Reset retries following a successful cross_validation call
                retries = 0

                # Store the model params and validation scores
                model_params_tested.append(model_params)
                val_scores.append(scores["r2"])  # type: ignore  # noqa: PGH003

                # Log the tuning iteration results
                logger.debug(
                    "tuning model: %s; iteration: %d/%d; mean (std) r2=%.3f (%.3f); "
                    "model_params: %s",
                    model_class.model_name(),
                    len(model_params_tested),
                    self.n_iter,
                    np.mean(scores["r2"]),
                    np.std(scores["r2"]),
                    str(model_params),
                )

            except Exception as e:
                # Increment retries following a failed tuning iteration and try again
                retries += 1
                logger.warning(
                    "Failed tuning iteration %d with model params: %s: %s",
                    len(model_params_tested),
                    json.dumps(model_params, default=str, separators=(",", ":")),
                    str(e),
                )
                # If many consecutive retries, log error and raise exception
                if retries > max_retries:
                    logger.error(
                        "Failed after %s with exception %s", max_retries, str(e)
                    )
                    msg = (
                        f"Failed to tune model {model_class.model_name()} after "
                        f"{max_retries} retries."
                    )
                    raise RuntimeError(msg) from e

        return val_scores, model_params_tested



import torch
from sklearn.preprocessing import PolynomialFeatures
from torch import nn

from autoemulate.core.device import TorchDeviceMixin
from autoemulate.core.types import DeviceLike, TensorLike
from autoemulate.data.utils import set_random_seed
from autoemulate.emulators.base import PyTorchBackend
from autoemulate.transforms.standardize import StandardizeTransform


class PolynomialRegression(PyTorchBackend):
    """
    PolynomialRegression emulator.

    Implements a linear model including all main effects, interactions,
    and quadratic terms.
    """

    def __init__(  # noqa: PLR0913
        self,
        x: TensorLike,
        y: TensorLike,
        standardize_x: bool = False,
        standardize_y: bool = False,
        degree: int = 2,
        lr: float = 0.1,
        epochs: int = 500,
        batch_size: int = 16,
        random_seed: int | None = None,
        device: DeviceLike | None = None,
        **kwargs,
    ):
        """Initialize a PolynomialRegression emulator.

        Parameters
        ----------
        x: TensorLike
            Input features.
        y: TensorLike
            Target values.
        standardize_x: bool
            Whether to standardize input features. Defaults to False.
        standardize_y: bool
            Whether to standardize target values. Defaults to False.
        degree: int
            Degree of the polynomial features to be generated. Defaults to 2.
        lr: float
            Learning rate for the optimizer. Defaults to 0.1.
        epochs: int
            Number of training epochs. Defaults to 500.
        batch_size: int
            Batch size for training. Defaults to 16.
        random_seed: int | None
            Random seed for reproducibility. Defaults to None.
        device: DeviceLike | None
            Device to run the model on. If None, uses the default device. Defaults to
            None.
        **kwargs: dict
            Additional keyword arguments.
        """
        super().__init__()
        TorchDeviceMixin.__init__(self, device=device)
        if random_seed is not None:
            set_random_seed(seed=random_seed)
        self.x_transform = StandardizeTransform() if standardize_x else None
        self.y_transform = StandardizeTransform() if standardize_y else None
        self.degree = degree
        self.lr = lr
        self.epochs = epochs
        self.batch_size = batch_size

        # Fit the polynomial feature transformer on the input data
        self.poly = PolynomialFeatures(degree=self.degree)
        x_np, _ = self._convert_to_numpy(x)
        x_poly = self.poly.fit_transform(x_np)
        self.n_outputs_ = y.shape[1] if y.ndim > 1 else 1
        self.linear = nn.Linear(x_poly.shape[1], self.n_outputs_, bias=False).to(
            self.device
        )
        self.optimizer = self.optimizer_cls(self.linear.parameters(), lr=self.lr)  # type: ignore[call-arg] since all optimizers include lr
        self.scheduler_setup(kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through for polynomial regression."""
        # Transform input using the fitted PolynomialFeatures
        x_np, _ = self._convert_to_numpy(x)
        x_poly = self.poly.transform(x_np)
        x_poly_tensor = torch.tensor(x_poly, dtype=torch.float32, device=self.device)
        return self.linear(x_poly_tensor)

    @staticmethod
    def is_multioutput() -> bool:
        """Polynomial regression supports multi-output."""
        return True

    @staticmethod
    def get_tune_params():
        """Return a dictionary of hyperparameters to tune."""
        scheduler_params = PolynomialRegression.scheduler_params()
        return {
            "lr": [1e-3, 1e-2, 1e-1, 2e-1],
            "epochs": [50, 100, 200, 500, 1000],
            "batch_size": [8, 16, 32],
            "scheduler_cls": scheduler_params["scheduler_cls"],
            "scheduler_kwargs": scheduler_params["scheduler_kwargs"],
        }



import logging

import torch
from torch import nn

from autoemulate.core.types import DeviceLike, TensorLike

SUPPORTED_DEVICES: list[str] = ["cpu", "mps", "cuda", "xpu"]


class TorchDeviceError(NotImplementedError):
    """Exception raised when the device is not implemented in torch."""

    def __init__(self, device: str):
        msg = f"Backend ({device}) not implemented."
        super().__init__(msg)


def get_torch_device(device: DeviceLike | None) -> torch.device:
    """
    Get the device returning the torch default device if None.

    Parameters
    ----------
    device: DeviceLike | None
        The device to get. If None, the default torch device is returned.

    Returns
    -------
    torch.device
        The device.

    Raises
    ------
    TorchDeviceError
        If the device is not a valid torch device.
    """
    if isinstance(device, torch.device):
        return device
    if device is None:
        return torch.get_default_device()
    if device in SUPPORTED_DEVICES:
        return torch.device(device)
    raise TorchDeviceError(device)


def move_tensors_to_device(
    *args: TensorLike, device: torch.device
) -> tuple[TensorLike, ...]:
    """
    Move the given tensor to the device.

    Parameters
    ----------
    *args: TensorLike
        The tensors to move.
    device: torch.device
        The device to move the tensors to.

    Returns
    -------
    tuple[TensorLike, ...]
        The tensors on the device.
    """
    return tuple(tensor.to(device) for tensor in args)


def check_torch_device_is_available(device: DeviceLike) -> bool:
    """
    Check if the given device type is available.

    Parameters
    ----------
    device: DeviceLike
        The device to check.

    Returns
    -------
    bool
        True if the device is available, False otherwise.

    Raises
    ------
    TorchDeviceError
        If the device is not a valid torch device.
    """
    if device == "cpu" or (
        isinstance(device, torch.device) and device.type == torch.device("cpu").type
    ):
        return True
    if device == "mps" or (
        isinstance(device, torch.device) and device.type == torch.device("mps").type
    ):
        return torch.backends.mps.is_available()
    if device == "cuda":
        return torch.cuda.is_available()
    if isinstance(device, torch.device) and device.type == "cuda":
        if device.index is not None:
            return device.index < torch.cuda.device_count()
        return True
    if device == "xpu" or (
        isinstance(device, torch.device) and device.type == torch.device("xpu").type
    ):
        return torch.xpu.is_available() if hasattr(torch, "xpu") else False
    raise TorchDeviceError(str(device))


def check_model_device(model: nn.Module, expected_device: str) -> bool:
    """
    Check if the model is on the expected device.

    Parameters
    ----------
    model: nn.Module
        The model to check.
    expected_device: str
        The expected device.

    Returns
    -------
    bool
        True if the model is on the expected device (ignoring device index), False
        otherwise.
    """
    return (
        str(next(model.parameters()).device).split(":")[0]
        == str(expected_device).split(":")[0]
    )


class TorchDeviceMixin:
    """
    Mixin class to add device management to a PyTorch model.

    Attributes
    ----------
    device: torch.device
        The device to use. If None, the default torch device is used.

    Raises
    ------
    TorchDeviceError
        If the device is not a valid torch device.
    """

    def __init__(self, device: DeviceLike | None = None, cpu_only: bool = False):
        # Warn if given device not CPU and cpu_only
        # TODO: check handling
        if cpu_only and (
            (isinstance(device, str) and device != "cpu")
            or (isinstance(device, torch.device) and torch.device("cpu") != device)
        ):
            msg = (
                f"The device ({device}) must be CPU for given model. Setting device as "
                "'cpu'."
            )
            # warnings.warn(msg, stacklevel=2)
            logging.warning(msg)

        self.device = get_torch_device(device)

        if not check_torch_device_is_available(self.device):
            raise TorchDeviceError(str(self.device))

    def _move_tensors_to_device(self, *args: TensorLike) -> tuple[TensorLike, ...]:
        """
        Move the given tensor to the device.

        Parameters
        ----------
        *args: TensorLike
            The tensors to move.

        Returns
        -------
        tuple[TensorLike, ...]
            The tensors on the device.
        """
        return move_tensors_to_device(*args, device=self.device)



import numpy as np
import torch
from scipy.integrate import solve_ivp

from autoemulate.core.types import NumpyLike, TensorLike
from autoemulate.simulations.base import Simulator


class Epidemic(Simulator):
    """Simulator of infectious disease spread (SIR)."""

    def __init__(
        self,
        parameters_range=None,
        output_names=None,
        log_level: str = "progress_bar",
    ):
        if parameters_range is None:
            parameters_range = {"beta": (0.1, 0.5), "gamma": (0.01, 0.2)}
        if output_names is None:
            output_names = ["infection_rate"]
        super().__init__(parameters_range, output_names, log_level)

    def _forward(self, x: TensorLike) -> TensorLike:
        """
        Simulate the epidemic using the SIR model.

        Parameters
        ----------
        x: TensorLike
            input parameter values to simulate [beta, gamma]:
            - `beta`: the transimission rate per day
            - `gamma`: the recovery rate per day

        Returns
        -------
        TensorLike
            Peak infection rate.
        """
        assert x.shape[0] == 1, (
            f"Simulator._forward expects a single input, got {x.shape[0]}"
        )
        y = simulate_epidemic(x.cpu().numpy()[0])
        # TODO (#537): update with default dtype
        return torch.tensor([y], dtype=torch.float32).view(-1, 1)


def simulate_epidemic(x: NumpyLike, N: int = 1000, I0: int = 1) -> float:
    """
    Simulate an epidemic using the SIR model.

    Parameters
    ----------
    x: NumpyLike
        The parameters of the SIR model. The first element is the transmission rate
        (beta) and the second element is the recovery rate (gamma).
    N: int
        The total population size. Defaults to 1000.
    I0: int
        The initial number of infected individuals. Defaults to 1.

    Returns
    -------
    peak_infection_rate: float
        The peak infection rate as a fraction of the total population.
    """
    # check inputs
    assert len(x) == 2
    assert N > 0
    assert I0 > 0
    assert I0 < N
    assert x[0] > 0
    assert x[1] > 0

    # unpack parameters
    beta = x[0]
    gamma = x[1]

    S0 = N - I0
    R0 = 0
    t_span = [0, 160]
    y0 = [S0, I0, R0]

    def sir_model(t, y, N, beta, gamma):  # noqa: ARG001
        S, I, R = y  # noqa: E741
        dSdt = -beta * S * I / N
        dIdt = beta * S * I / N - gamma * I
        dRdt = gamma * I
        return [dSdt, dIdt, dRdt]

    t_eval = np.linspace(
        t_span[0], t_span[1], 160
    )  # Evaluate each day within the time span
    sol = solve_ivp(sir_model, t_span, y0, args=(N, beta, gamma), t_eval=t_eval)

    S, I, R = sol.y  # noqa: E741
    I_max = np.max(I)

    # peak infection rate
    return I_max / N



from pathlib import Path

import pandas as pd
import torch
from sklearn.model_selection import train_test_split

from autoemulate.core.types import TensorLike

# Data is stored in the same directory
data_dir = Path(__file__).parent


def fetch_data(
    dataset: str, split: bool = False, test_size: float = 0.2, random_state: int = 42
) -> list[TensorLike]:
    """
    Fetch a dataset by name.

    Parameters
    ----------
    dataset: str
        Dataset to load. Can be any of the following strings:
        - **cardiac1**: ionic atrial cell model data from LH sampling.
        - **cardiac2**: isotonic contraction ventricular cell model, no LH sampling.
        - **cardiac3**: CircAdapt: four-chamber pressure and volume CircAdapt ODE model
                        from LH sampling.
        - **cardiac4**: four chamber: 3D-0D four-chamber electromechanics model to
                        predict pressure and volume biomarkers for cardiac function.
        - **cardiac5**: passive mechanics: inflated volumes and mean atrial and
                        ventricular fiber strains for a passive inflation.
        - **cardiac6**: tissue electrophysiology: predict total atrial and ventricular
                        activation times with an Eikonal model.
        - **climate1**: GENIE model: predict climate variables SAT, ACC, VEGC, SOILC,
                        MAXPMOC, OCN_O2, fCaCO3, SIAREA_S.
        - **engineering1**: Cantilever truss simulation.
        - **reactiondiffusion1**: Reaction-diffusion system simulated data. Spatial
                        snapshots in correspondence for different values of reaction
                        and  ddiffusion parameters. (50 snapshots in correspondence of
                        50 different parameter sets at time T=10).
    split: bool
        Whether to split the data into training and testing sets. Default is False.
    test_size: float
        The proportion of the dataset to include in the test split. Default is 0.2.
    random_state: int
        Controls the shuffling applied to the data before applying the split. Default
        is 42.

    Returns
    -------
    list[TensorLike]
        Tensors of simulation parameters / inputs and simulation outputs. If `split`
        is True, returns [x_train, x_test, y_train, y_test].
    """
    data_dir_dataset = data_dir / dataset / "processed"
    X = pd.read_csv(data_dir_dataset / "parameters.csv").to_numpy()
    Y = pd.read_csv(data_dir_dataset / "outputs.csv").to_numpy()

    if split:
        x_train, x_test, y_train, y_test = train_test_split(
            X, Y, test_size=test_size, random_state=random_state
        )
        return [
            torch.tensor(x_train),
            torch.tensor(x_test),
            torch.tensor(y_train),
            torch.tensor(y_test),
        ]
    return [torch.tensor(X), torch.tensor(Y)]
